/**
	@file	avrfat16.c
	@brief	Fat16 Functions
	@author	Nick Lott 
	@date	September 2004

	Copyright (C) 2004 Nick Lott <brokentoaster@sf.net>

	Minor changes for a digital picture frame by M.Bode 2007

	This is a simple implementation of the FAT16 file system. It is designed
	to be small for use with MP3 players and MMC cards. Currently it is 
	readonly.NOTE: The code acknowledges only the first partition on the drive.
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software Foundation, 
	Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	

	Target(s)...: ATmega169

	Compiler....: AVR-GCC 3.3.1; avr-libc 1.0

**/

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>


#include "types.h"
#include "avrfat16.h"
#include "utils.h"


//void FAT_PrintRecord(uint8 *pRecord);

/**
*   FAT_initFat16
*	@brief		read MBR and bootsector and set global fat variables
*	@return 	non zero for error.
**/
uint8 FAT_initFat16(void) 
{
	uint8 	result,type,lba_begin,sectors_per_cluster,number_of_fats;
	uint16	bytes_per_sector,number_of_reserved_sectors,BPB_RootEntCnt;
	uint16	sectors_per_fat;
	
	result = FAT_read(0);
	
	if (result)
		return result; // abort on read error
	
	if (FAT_buffer[510] != 0x55 || FAT_buffer[511] != 0xAA)
		return 0x1A; // abort if signature bytes not found in MBR
		
	type = FAT_buffer[PARTITION_START+4]; // get type code for partition one
	
	// TODO: check here for valid MMC types ie FAT16 = 0x
	
	// get start of FAT partition.
	lba_begin = FAT_buffer[PARTITION_START+11];
	lba_begin <<=8;	
	lba_begin += FAT_buffer[PARTITION_START+10];
	lba_begin <<=8;	
	lba_begin += FAT_buffer[PARTITION_START+9];
	lba_begin <<=8;	
	lba_begin += FAT_buffer[PARTITION_START+8];
	
	
	// read FAT Volume ID
	result = FAT_read(lba_begin);
	if (result) return result; // abort on read error
	
	//if (FAT_buffer[510] != 0x55 || FAT_buffer[511] != 0xAA)
	//	return 0xAA; // abort if signature bytes not found in MBR

	
	sectors_per_cluster= FAT_buffer[0x0D];
	
	
	number_of_reserved_sectors = FAT_buffer[ 0x0f];
	number_of_reserved_sectors <<=8;
	number_of_reserved_sectors += FAT_buffer[ 0x0e];
	
	number_of_fats = FAT_buffer[0x10];
	
	bytes_per_sector = FAT_buffer[0x0c];
	bytes_per_sector <<=8;
	bytes_per_sector += FAT_buffer[0x0b];

	// BPB_FATSz16
	sectors_per_fat = FAT_buffer[0x17];
	sectors_per_fat <<=8;
	sectors_per_fat += FAT_buffer[0x16];
	
	//BPB_RootEntCnt
	BPB_RootEntCnt = FAT_buffer[0x12];
	BPB_RootEntCnt <<=8;
	BPB_RootEntCnt += FAT_buffer[0x11];

	
	/*
	FOR FAT32....
	sectors_per_fat = FAT_buffer[0x27];
	sectors_per_fat <<=8;
	sectors_per_fat += FAT_buffer[0x26];
	sectors_per_fat <<=8;
	sectors_per_fat += FAT_buffer[0x25];
	sectors_per_fat <<=8;
	sectors_per_fat += FAT_buffer[0x24];
	
	root_director_first_cluster = FAT_buffer[0x2f];
	root_director_first_cluster <<=8;
	root_director_first_cluster += FAT_buffer[0x2e];
	root_director_first_cluster <<=8;
	root_director_first_cluster += FAT_buffer[0x2d];
	root_director_first_cluster <<=8;
	root_director_first_cluster += FAT_buffer[0x2c];
	*/
	
	// note: FAT16 clusters start after root directory.
	FAT16_fat_begin_lba = lba_begin + number_of_reserved_sectors;
	FAT16_root_dir_first_cluster = FAT16_fat_begin_lba + ( number_of_fats * sectors_per_fat);
	FAT16_cluster_begin_lba = FAT16_root_dir_first_cluster+
		((BPB_RootEntCnt*32) + (bytes_per_sector-1))/bytes_per_sector;
	//	((BPB_RootEntCnt*32) + (bytes_per_sector))/bytes_per_sector;
	FAT16_sectors_per_cluster = sectors_per_cluster;
	
	FAT16_dir_first_cluster=FAT16_root_dir_first_cluster;
	FAT16_parent_dir_first_cluster = FAT16_root_dir_first_cluster;
	
	return 0;
}

/**
*   FAT_get_label
*	@brief		copys the volume label
*	@param		destentaion for label
*	@return 	non zero for error.
**/
uint8 FAT_get_label(uint8 label[])
{
	uint8 result,i,j;
	
	// read root dir
	result = FAT_read(FAT16_root_dir_first_cluster);
	if (result) return result; // abort on fail
	
	// find entry with ATTR_VOLUME_ID
	for (i=0;i<512/32;i++){
		if (FAT_buffer[(i*32)+11] == FILE_ATTR_VOLUME_ID){
			for (j=0;j<11;j++)
				label[j]=FAT_buffer[(i*32)+j];
			label[11]=0;
			return 0;// we have a volume
		}
	}
	
	return 1;
	
	
}


/**
*   FAT_readCluster
*	@brief		Reads a sector from a cluster into the buffer
*	@param		cluster		Clusterto read from
*	@param		sector_offset	Sector to read from within cluster
*	@return		error code from FAT_read()
**/
uint8 FAT_readCluster(uint32 cluster, uint8 sector_offset) 
{
	uint8	result=0;
	uint32 lba_addr;
	
	// calculate sector to read
	lba_addr = FAT16_cluster_begin_lba + ((cluster-2) * FAT16_sectors_per_cluster);
	lba_addr += sector_offset;
	
	result = FAT_read(lba_addr);
	
	return result;
}


/**
*   FAT_NextCluster
*	@brief		looks up a cluster in fat
*	@param		cluster 	current cluster
*	@return		next cluster
*	@return 	0 for error 
*	@return  	0xffffff for end of file.
**/
uint32 FAT_NextCluster(uint32 cluster)
{
	uint8 result=0;
	uint32 sector,position,next;
	
	if (cluster ==0 ) cluster = 2; // check for reserved cluster mishap ;)
	
	//sector = cluster / 128; 
	sector = cluster / 256; // for fat 16
	
	result = FAT_read(FAT16_fat_begin_lba+sector);
	if (result) return 0;
	
	//position = (cluster - (sector * 128)) * 4; for fat 32
	position = (cluster - (sector * 256)) * 2; // for fat 16
	
	/*
	// read 4 bytes (FAT32)
	next = FAT_buffer[position+3];
	next <<= 8;
	next += FAT_buffer[position+2];
	next <<= 8;
	next = FAT_buffer[position+1];
	next <<= 8;
	next += FAT_buffer[position];
	*/
	
	// read 2 Bytes (FAT16)
	next = FAT_buffer[position+1];
	next <<= 8;
	next += FAT_buffer[position];

	// mask off last 4 bits (FAT32)
	// next &= 0x0ffffff;
	
	if (next == 0xffff) next = 0xffffffff;
	
	return next;
}

/**
*   FAT_read
 *	@param		filenumber 32bit uInt pointing to file of interst.
 *	@param		dir_first_entry 32bit uInt pointing to first cluster of directory
 *	@return		error code
 *
 *	Read any directory of the disk and obtain details 
 *	about FAT entry number <filenumber>
 *
 *	error codes:
 *	0x00	mp3file found
 *	0x02	end of directory
 *	0x59	deleted file
 *
 **/

uint8	FAT_readFile(uint32 filenumber, uint32 dir_first_cluster)
{
	uint8 result,i,lfn;
	uint8 record,lfn_record,nmeChkSm,lfnchksm;
	uint8 attrib;
	uint16 lfn_entry,fat_entry; // position in buffer of entry 0-512
	uint32 sector,lfn_sector;
//	uint16 cluster;
//	uint32 size;

	record = filenumber & 0x0F;
	sector = filenumber >> 4;
	fat_entry =	record << 5;
	
	// read the root dir
	result = FAT_read(dir_first_cluster+sector);
	if (result) return result; // abort on non-zero reply 

	// check firstByte
	if (FAT_buffer[fat_entry] == 0x00){ // end of directory
		FAT16_entryMAX = filenumber;
		return 2;
	}
	else if (FAT_buffer[fat_entry] != 0xe5){ // not used (aka deleted)
		
		// exit on . and ..
		if (FAT_buffer[fat_entry]=='.')
		{
			return 0x59;
		}
		
		// get the attrib byte
		attrib = FAT_buffer[fat_entry+FAT16_LDIR_Attr];
		
		if (attrib == FILE_TYPE_FILE || attrib== FILE_TYPE_DIR ){ // entry is normal 8.3 entry
			
			// construct short filename string
			for (i=0;i<12;i++){ 
				FAT_scratch[i] = FAT_buffer[fat_entry+i];
			}
			FAT_scratch[11] = '/';
			if (attrib == FILE_TYPE_FILE ){
				FAT_scratch[8] = '.';
				FAT_scratch[9] = FAT_buffer[fat_entry+8];
				FAT_scratch[10] = FAT_buffer[fat_entry+9];
				FAT_scratch[11] = FAT_buffer[fat_entry+10];
			}
			FAT_scratch[12] = 0x00;
				
			// accept only *.bmp files
			if(attrib== FILE_TYPE_DIR ||
			   ( FAT_scratch[9] =='B' && 
				FAT_scratch[10] =='M' && 
				FAT_scratch[11] =='P')){

				
				// get Cluster 
				FAT_scratch[13] = FAT_buffer[(fat_entry)+0x15];
				FAT_scratch[14] = FAT_buffer[(fat_entry)+0x14];
				FAT_scratch[15] = FAT_buffer[(fat_entry)+0x1B];
				FAT_scratch[16] = FAT_buffer[(fat_entry)+0x1A];
				
				//get fileSize
				FAT_scratch[17] = FAT_buffer[(fat_entry)+0x1f];
				FAT_scratch[18] = FAT_buffer[(fat_entry)+0x1e];
				FAT_scratch[19] = FAT_buffer[(fat_entry)+0x1d];
				FAT_scratch[20] = FAT_buffer[(fat_entry)+0x1c];
				
				//
				// Do Long Filename entries....
				
				//checksum of name for long entries..
				nmeChkSm=FAT_ChkSum(FAT_buffer+fat_entry);
				
				lfn_record = record;
				lfn_sector = sector;
				lfn_entry =	lfn_record << 5;
				
				do{
					// look at previous FAT entry
					if (lfn_record > 0){
						lfn_record--;
					}else{
						if (lfn_sector > 0) {
							lfn_sector--;
						
							// read in new sector
							result = FAT_read(FAT16_root_dir_first_cluster+lfn_sector);
							if (result){ // abort on non-zero reply 
								break;
							}
						
							lfn_record = 0x0F;// 15 records per sector FAT16
							
						}else{ // force exit if we run out of sectors
							break;
						}
					}
					lfn_entry =	lfn_record << 5;
					
					// dump record for debugging
					//FAT_PrintRecord(FAT_buffer+lfn_entry);EOL();
					
					// get the attrib byte
					attrib = FAT_buffer[lfn_entry+FAT16_LDIR_Attr];
					
					//exit if no LFN entry found
					if (attrib != FILE_ATTR_LFN_TEXT){
						break;
					}
					
					//get the chksm
					lfnchksm = FAT_buffer[lfn_entry+FAT16_LDIR_Chksum];
		
					// exit if ChkSum doesn't Match
					if (lfnchksm != nmeChkSm){
						break;
					}
					
					//look at order, select lfn section
					lfn = ((FAT_buffer[lfn_entry+FAT16_LDIR_Ord] & FILE_ATTR_LFN_MASK)-1) *13;

               if (lfn < LFN_BUFFER_LENGTH-13)  // Only allow 208 byte max LFNs
               {
   					// note: only look at lower byte of unicode text.
   					FAT16_longfilename[13+lfn] = 0x00;
   					FAT16_longfilename[12+lfn] = FAT_buffer[lfn_entry+0x1e];
   					FAT16_longfilename[11+lfn] = FAT_buffer[lfn_entry+0x1c];
   					FAT16_longfilename[10+lfn] = FAT_buffer[lfn_entry+0x18];
   					FAT16_longfilename[9+lfn] = FAT_buffer[lfn_entry+0x16];
   					FAT16_longfilename[8+lfn] = FAT_buffer[lfn_entry+0x14];
   					FAT16_longfilename[7+lfn] = FAT_buffer[lfn_entry+0x12];
   					FAT16_longfilename[6+lfn] = FAT_buffer[lfn_entry+0x10];
   					FAT16_longfilename[5+lfn] = FAT_buffer[lfn_entry+0x0e];
   					FAT16_longfilename[4+lfn] = FAT_buffer[lfn_entry+0x09];
   					FAT16_longfilename[3+lfn] = FAT_buffer[lfn_entry+0x07];
   					FAT16_longfilename[2+lfn] = FAT_buffer[lfn_entry+0x05];
   					FAT16_longfilename[1+lfn] = FAT_buffer[lfn_entry+0x03];
   					FAT16_longfilename[lfn] = FAT_buffer[lfn_entry+0x01];
               }

				// test to see if we need to keep looking
				}while(FAT_buffer[lfn_entry] < FILE_ATTR_LFN_MASK);	// not last entry

				if (lfn_sector != sector) {// re read original FAT sector if changed
					result = FAT_read(FAT16_root_dir_first_cluster+sector);
					if (result) return result; // abort on non-zero reply 
				}
				
				return 0x00;
			}
		}			
	}
	
	return 0x57;
}

/**
*   FAT_readRoot
 *	@param		filenumber 32bit uInt pointing to file of interst.
 *	@return		error code
 *
 *	Read the root directory of the disk and obtain details 
 *	about FAT entry number <filenumber>
 **/
uint8	FAT_readRoot(uint32 filenumber)
{
	return FAT_readFile(filenumber,FAT16_root_dir_first_cluster);
}


/**
*	GetNextSong
 *	@return new filenumber or 0=err
 *
 **/
uint32 FAT_getNextSong(uint32 filenumber,uint32 directory){
	uint16 i;
	uint8 result=0;
	//uint32 nextfile;
	
	if (filenumber < FAT16_entryMAX){
		filenumber++;
		
		// find next valid FAT entry
	
		for(i=0;i<1023;i++)
		{
			result = FAT_readFile(filenumber++,directory);
			if(result==0)
			{
				filenumber--;
				return filenumber;
			}
			if(result == 0x59 || result == 2)	break;	
		}	
	}
	return 0;
	
}


//----------------------------------------------------------------------------- 
// FAT_ChkSum() 
// Returnsanunsignedbytechecksumcomputedonanunsignedbyte 
// array. Thearraymustbe11byteslongandisassumedtocontain 
// anamestoredintheformatofaMS-DOSdirectoryentry. 
// Passed: pFcbName Pointertoanunsignedbytearrayassumedtobe 
// 11byteslong. 
// Returns:Sum An8-bitunsignedchecksumofthearraypointed 
// tobypFcbName. 
//------------------------------------------------------------------------------ 
uint8 FAT_ChkSum(uint8 *pFcbName) 
{ 
	uint8 FcbNameLen; 
	uint8 Sum;
	
	//PRINT(">> ");
	Sum=0; 
	for(FcbNameLen=11;FcbNameLen!=0;FcbNameLen--){ 
		//UART_SendByte(*pFcbName);
		//NOTE:The operation is an unsignedchar rotate right 
		Sum=((Sum&1)?0x80:0)+(Sum>>1)+*pFcbName++; 

	} 
	//PRINT(" ");
	//UART_Printfu08(Sum);
	//EOL();
	return(Sum); 
} 
   
/**
*	FAT_scanDir
 *	@param	uint32 pointing to director start cluster
 *	@return	number of files found
 *
 *	scan a directory for songs, set max, min and currentdir variables
 **/
uint8 FAT_scanDir(uint32 directory)
{
	uint8 files=0;
	uint32 file=-1,nextfile;
	
	// exit if already in this directory.
	//if (directory==FAT16_dir_first_cluster) return -1;
	
	// get first good file
	FAT16_entryMIN = FAT_getNextSong(0,directory);

	//scan root directory and get max File Entry.
	nextfile = FAT16_entryMIN;
	while ( nextfile ){
		file = nextfile;
		files++;
		nextfile = FAT_getNextSong(file,directory);
	}

	FAT16_entryMAX =file; // set to last valid MP3
	
	if (files){ // then set working directory
		FAT16_parent_dir_first_cluster = FAT16_dir_first_cluster;
		FAT16_dir_first_cluster = directory;
	}
	
	return files;
}
